#ifndef HPP__DATABASE
#define HPP__DATABASE

#include <cstddef> /* size_t */
#include <wx/thread.h>  /* wxCriticalSection */

#if !wxUSE_THREADS
    #error "This sample requires thread support!"
#endif

#include "net_types_basic.hpp" /* uintMAC, uintIP, uintPort */

#include "net_types_containers.hpp"  /* MAC_Data, IP_Data, Port_Data */

#include "GUI_Dialog_Main.hpp"  /* For the "DialogMain" type */

#include "reusable/superarray.hpp"  /* SuperArray< T, len > */

#include "frame_analysis.hpp"  /* FrameAnalysis class */

/* An object of the Sniffer_Database class
   is THREAD-SAFE, it deals internally with
   locking shared resources whenever any
   thread tries to access it.
       It achieves this by locking resources
   in the "find_in_container_or_add" method. */

#include <utility>  /* pair */

class Sniffer_Database {
public:

    typedef SuperArray<MAC_Data,const_mac_max_qty> ContainerType_MAC;

private:

    wxCriticalSection cs;

    ContainerType_MAC mac_data_container;

    IntType_Qty_IP_Ever_Logged qty_total_ips_logged;

public:

    class Locker_Readonly_Accessor {
    private:

        Sniffer_Database &sdb;

    public:

        Locker_Readonly_Accessor(Sniffer_Database &arg_sdb) : sdb(arg_sdb)
        {
            sdb.cs.Enter();
        }

        ~Locker_Readonly_Accessor()
        {
            sdb.cs.Leave();
        }

        ContainerType_MAC const &Get_MAC_Container()
        {
            return sdb.mac_data_container;
        }

        IntType_Qty_IP_Ever_Logged GetQty_IP_Ever_Logged()
        {
            return sdb.qty_total_ips_logged;
        }
    };


public:

    /* An object of this class is about 5 megabytes in size so you'd wanna
       create it on the heap (i.e. new) */

    Sniffer_Database()
        : cs(), mac_data_container(), qty_total_ips_logged(0) {}

private:

    /* The following method locks automatically when writing,
       however it does NOT lock when reading. Therefore it should
       only be invoked by the Sniffer Thread. */

    template<class Container, typename T>
    std::pair<typename Container::iterator,bool> find_in_container_or_add(Container &c, T const &val);

    /* This function is called internally by the public function that follows */

    void SubmitData_For_One_Direction(uintMAC mac, uintIP ip, uintPort port);

public:

#if 0
    MAC_Data const *GetAddressOf_MAC_Data(size_t const i)
    {
        return this->mac_data_container.begin() + i;
            /* This is thread-safe if begin is REENTRANT
               and the containers begin position never
               changes (be careful with magical containers
               like vector that might reallocate memory
               elsewhere). */
    }
#endif

    /* The following method locks automatically when writing,
       however it does NOT lock when reading. Therefore it should
       only be invoked by the Sniffer Thread. */

    void SubmitFrameAnalysis(FrameAnalysis const &fa);

    /* The following method does NEVER locks, and therefore it should
       only be called by the main GUI thread when all other
       threads are dead. */

    void Reset_Database()
    {
        mac_data_container.clear();

        qty_total_ips_logged = 0;
    }

private:

    /* ------ Prevent copy-construction and assignment------ */
    Sniffer_Database(Sniffer_Database const &)
    {

    }

    Sniffer_Database &operator=(Sniffer_Database const &);
    /* ----------------------------------------------------- */

};

extern Sniffer_Database *g_p_sdb;

#endif

