#include "parser.hpp"

#include <cassert>
#include <cstring> /* strlen */
#include <cctype>  /* isdigit */
using std::strlen;
using std::isdigit;

#include "master_specs_for_dynamo_program.hpp"  /* const_max_qty_nets_to_arp */

//#include <wx/msgdlg.h>

wxString MakeStringFromMAC(uintMAC const mac)
{
    assert(mac <= const_mac_max);

    wxString str;

    str.Printf(wxT("%02X:%02X:%02X:%02X:%02X:%02X"),
    (unsigned)(mac >> 40),
    (unsigned)(mac >> 32 & 0xFFu),
    (unsigned)(mac >> 24 & 0xFFu),
    (unsigned)(mac >> 16 & 0xFFu),
    (unsigned)(mac >> 8 & 0xFFu),
    (unsigned)(mac & 0xFFu));

    return str;
}

wxString MakeStringFromIP(uintIP const ip)
{
    assert(ip <= const_ip_max);

    wxString str;

    str.Printf(wxT("%u.%u.%u.%u"),
    (unsigned)(ip >> 24),
    (unsigned)(ip >> 16 & 0xFFu),
    (unsigned)(ip >> 8 & 0xFFu),
    (unsigned)(ip & 0xFFu));

    return str;
}

bool ParseDecimalNumber_fast8(char const *const p,uint_fast8_t const max_chars,uint_fast8_t &sum)
{
	/* Must allow for empty strings because of ParseNetworkAddr */

    assert(p != 0);
    assert(max_chars != 0);
    assert(max_chars <= 3); /* Because 255 is the highest number! */

    if (!p[0]) /* If string is empty, don't put !p[1] in here!!! */
        return false;

    size_t const len = strlen(p);

    if (len > max_chars)
		return false;

    /* Get a pointer to the last digit */
    char const *q = p + (len - 1);

    uint_fast16_t multiplier = 1;
    sum = 0;

    for ( ; ; --q)
    {
        if ( !isdigit((char unsigned)*q) )
            return false;

        sum += (*q - '0') * multiplier;

        multiplier *= 10;

        if (p == q)
            break;
    }

    return true;
}

bool ParseDecimalNumber_fast16(char const *const p,uint_fast8_t const max_chars,uint_fast16_t &sum)
{
	/* Must allow for empty strings because of ParseNetworkAddr */

    assert(p != 0);
    assert(max_chars != 0);
    assert(max_chars <= 5); /* Because 65535 is the highest number! */

    if (!p[0]) /* If string is empty, don't put !p[1] in here!!! */
        return false;

    size_t const len = strlen(p);

    if (len > max_chars)
		return false;

    /* Get a pointer to the last digit */
    char const *q = p + (len - 1);

    uintIP multiplier = 1;
    sum = 0;

    for ( ; ; --q)
    {
        if ( !isdigit((char unsigned)*q) )
            return false;

        sum += (*q - '0') * multiplier;

        multiplier *= 10;

        if (p == q)
            break;
    }

    return true;
}

bool ParseIP(char const *const p, uintIP &ip) /* e.g. 134.3.67.189 */
{
    /* Can't have a null pointer */
    assert(p != 0);

    if (!p[0] || !p[1]) /* If string is empty or really small */
        return false;

    /* Get a pointer to the last char */
    char const *q = p + (strlen(p) - 1);

    /* Record the final IP address */
    ip = 0;

    /* Keep track of which octet is currently being processed */
    uint_fast8_t i_octet;

    /* Now use a loop to process 4 octets */
    for (i_octet = 0; i_octet != 4; ++i_octet)
    {
        uint_fast16_t octet_sum = 0; /* 16-Bit in case we get 999 */

        /* Each octet must start with a decimal digit */
        if ( !isdigit(*q) )
            return false;

#define ADVANCE(m)                                          \
        if ( !isdigit(*q) )                                 \
            goto Encountered_Non_Digit;                     \
                                                            \
        octet_sum += (uint_fast16_t)(m) * (*q - '0');       \
                                                            \
        if (p == q)                                         \
        {                                                   \
            ip |= (uintIP)octet_sum << 24;           \
            break;                                          \
        }                                                   \
                                                            \
        --q;

        ADVANCE(1);
        ADVANCE(10);
        ADVANCE(100);

        if (octet_sum > 255u) /* Could be 999, that's why we use 16-Bit instead of 8-Bit */
            return false;

        /* No need to check for 4 digits in a row, it will be caught below */

      Encountered_Non_Digit:

        /* The only allowable non-digit is '.' */
        if (*q != '.')
            return false;

        /* The IP address can't start with a non-digit */
        if (p == q)
            return false;

        /* Everything seems fine, OR into the IP address */
        ip |= ((uintIP)octet_sum << (8u * i_octet));

        /* Point the pointer to the least significant digit of the next octet */
        --q;
    }

    if (i_octet != 3) /* If this is 4, then p != q, so it's a bad IP */
        return false;

    //wxMessageBox(MakeStringFromIP(ip));
    return true;
}

static bool IsNetAddrClean(NetworkAddress const na)
{
    assert(na.qty_net_bits >= 1 && na.qty_net_bits <= 30);

    uintIP const netmask = (const_ip_max << (32 - na.qty_net_bits)) & const_ip_max;
                   /* Need to AND it in case we have big ints on some systems */

    return (na.addr & netmask) == na.addr;
}

bool ParseNetAddr(char *str,SuperArray<NetworkAddress,const_max_qty_nets_to_arp> &container)
{
    NetworkAddress netaddr;

    assert(str != 0);

    if (!str[0] || !str[1]) /* If empty string (or really small) */
        return false;

    char *p = strchr(str,'/');

	if (!p)
	{
	    //wxMessageBox(wxT("Couldn't find forward slash"));
	    return false;
	}

	*p = '\0'; /* Replace the forward slash with a null terminator */

	if ( !ParseIP(str,netaddr.addr) )
	{
        //wxMessageBox(wxT("Couldn't parse the IP"));
	    return false;
	}


    /* ---- Now time to parse the network mask ---- */

	++p;   /* Go past the forward slash to the first digit of the mask */

	/* Now 'p' should be a string number between 1 and 30 */

	if ( !ParseDecimalNumber_fast8(p,2,netaddr.qty_net_bits) )
	{
	    //wxMessageBox(wxT("Couldn't parse the network mask"));
	    return false;
	}

	if (netaddr.qty_net_bits < 1 || netaddr.qty_net_bits > 30)
	{
        //wxMessageBox(wxT("Network mask out of range!"));
		return false;
	}

    if ( !IsNetAddrClean(netaddr) ) /* e.g. 192.168.1.1/24 (should be 192.168.1.0/24) */
    {
        //wxMessageBox(wxT("Network address isn't clean"));
        return false;
    }

    try
    {
        container.push_back(netaddr);
        return true;
    }
    catch (...)
    {
        return false;
    }
}

bool GetNetworksFromString(wxString wxstr, SuperArray<NetworkAddress,const_max_qty_nets_to_arp> &container)
{
    /* Takes into account the max amount of networks to arp */

    container.clear();

    //wxMessageBox(wxstr);

    /* ---- First turn it into a C string ---- */

    char str[500];

    strncpy( str, wxstr.char_str(), sizeof(str) - 2);

    wxString tmp( (char const *)str, wxConvUTF8 );

    //wxMessageBox(tmp);

    /* ---------------------------------------- */

    for ( char *pstart = str ; ; )
    {
        if ( !*pstart )
        {
            /* Should never ever reach here */

            return false;
        }

        char *pdivider = strchr(pstart,' ');

        if ( !pdivider ) /* There's no "space" at all in the pstart string */
        {
            /* Must be the last address */

            return ParseNetAddr(pstart, container);
        }

        /* ---- We reach here if pdivider points to a space ---- */

        *pdivider = '\0'; /* Replace the space with a null terminator */

        if ( !ParseNetAddr(pstart,container) ) /* If network string is dodgy */
            return false;

        if ( const_max_qty_nets_to_arp == container.size() )
            return true; /* We've gotten the maximum amount of networks */

        pstart = pdivider+1; /* Move to after the space */
    }
}

char const *Port_To_Protocol_Name(uint_fast16_t const port)
{
    if (port > 1024)
        return 0;

    /* 1025 because it's Port 0 through port 1024 */
    static char const *ports[1025] = { 0 /* Don't remove this 1st value */,

#include "port_names.txt"

    };

    return ports[port];
}
