#ifdef __cplusplus
#error You must use a C compiler to compile this file (not a C++ compiler)
#endif

#include "rawsock.h"

#include <string.h>    /* strncpy */
#include <memory.h>    /* memset */
#include <stddef.h>    /* size_t */
#include <assert.h>    /* assert */

#ifndef VJC_LIB_RAW_SOCK_UNDERLYING_API

#    error "Must specify underlying API by defining VJC_LIB_RAW_SOCK_UNDERLYING_API"

#elif VJC_LIB_RAW_SOCK_UNDERLYING_API == 2 /* pcap */

#   include "rawsock_pcap.c.txt"

#elif VJC_LIB_RAW_SOCK_UNDERLYING_API == 0 /* This is the dummy */

RawEtherSockHandle OpenRawEtherSock(char const *const nic)
{
    RawEtherSockHandle h = {1};  /* Always succeed */

    return h;
}

int RecvRawEtherFrame(RawEtherSockHandle const h,FrameInfo *const p)
{
    return 0; /* Couldn't get a frame */
}

void SendRawEtherFrame(RawEtherSockHandle const h,uint8_t const *const p,uint_fast16_t const len)
{
    /* Return immediately as if frame was sent */
    return;
}

void CloseRawEtherSock(RawEtherSockHandle const h)
{
    /* Return immediately as if socket was closed */
    return;
}

char const *GetListOfInterfaces(uint_fast8_t const index)
{
    return 0;
}

#elif VJC_LIB_RAW_SOCK_UNDERLYING_API == 1 /* Native Linux API */

#include <sys/socket.h>        /* socket, bind, write, close */
#include <sys/ioctl.h>         /* ioctl, SIOCGIFINDEX */
#include <net/if.h>            /* ifreq, IFNAMSIZ */
#include <linux/if_ether.h>    /* ETH_P_ALL */
#include <netpacket/packet.h>  /* sockaddr_ll */
#include <netinet/in.h>        /* htons */
#include <unistd.h>            /* I don't know where "close" is declared */
#include <fcntl.h>

static uint_fast8_t SetSocketNonBlocking(int const s)
{
    int const flags = fcntl(s, F_GETFL, 0);

    /* Set socket to non-blocking */

    if ( flags < 0 )
    {
        return 0;
    }

    if ( fcntl(s, F_SETFL, flags | O_NONBLOCK) < 0 )
    {
        return 0;
    }

    return 1;
}


RawEtherSockHandle OpenRawEtherSock(char const *const nic)
{
    struct sockaddr_ll sll = {0};
    struct ifreq ifr = {0};

    RawEtherSockHandle h;

    /* First create the raw socket */
    h.s = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));

    if (-1 == h.s)
        return h;

    /* Copy the name of the NIC into
       this special structure          */

    strncpy(ifr.ifr_name, nic, IFNAMSIZ);

    /* Now get the index for the NIC
       (this is a special number that
        identifies the NIC to the
        operating system)             */

    if( -1 == ioctl(h.s, SIOCGIFINDEX, &ifr) )
    {
        /* Unable to obtain NIC index */
        close(h.s);
        h.s = -1;
        return h;
    }

    /* Bind our raw socket to the NIC */

    sll.sll_family = AF_PACKET;
    sll.sll_ifindex = ifr.ifr_ifindex;
    sll.sll_protocol = htons(ETH_P_ALL);

    if( -1 == bind(h.s,(struct sockaddr *)&sll,sizeof sll) )
    {
        /* Unable to bind raw socket to NIC */
        close(h.s);
        h.s = -1;
        return h;
    }

    /* OK, socket is open, now set it non-blocking */


    if( !SetSocketNonBlocking(h.s) )
    {
        /* Unable to set socket as non-blocking! */
        close(h.s);
        h.s = -1;
        return h;
    }

    return h;
}

int RecvRawEtherFrame(RawEtherSockHandle const h,FrameInfo *const pc)
{
    ssize_t retval = recv(h.s,pc->buf,sizeof pc->buf,MSG_DONTWAIT);

    if (retval < 1)
    {
        /* -1 means error occurred */

        /* 0 means there's no data in the buffer! */

        return 0;
    }

    pc->len = retval;

    return 1;
}

void SendRawEtherFrame(RawEtherSockHandle const h,uint8_t const *const p,uint_fast16_t const len)
{
    send(h.s,p,len,0);
}

void CloseRawEtherSock(RawEtherSockHandle const h)
{
    close(h.s);
}

char const *GetListOfInterfaces(uint_fast8_t const index)
{

    switch (index)
    {
    case 0: return "eth0";
    case 1: return "eth1";
    case 2: return "eth2";
    case 3: return "wlan0";
    case 4: return "wlan1";
    case 5: return "wlan8";

    default: return 0;
    }
}

#else

#   error "Unrecognised value for VJC_LIB_RAW_SOCK_UNDERLYING_API"

#endif
